#ifndef _SMP_INHERIT
#define _SMP_INHERIT


/**
*  List terminator. This does not need to be used in most cases.
*
*  SubTypeList already defaults its tail to this type
*/
struct EndTypeList
{
    //empty
};

/**
* Part of a lisp style list that contains one template type name.
* the template type must have only one explicit parameter of type
* 'class'
*
* This is useful for passing types to GenerateSubClass.
*
* Normal usage is:
* SubTypeList< FirstTypeName, 
          SubTypeList< SecondTypeName,
               SubTypeList<ThirdTypeName, 
                  SubTypeList<LastTypeName> > > >
*/
template <
    template<class>class Type, //template type of this node
    class Tail_ = EndTypeList  // rest of the list or end node
         >
struct SubTypeList
{
    typedef Tail_ Tail;

    template <class ChildType>
    struct UseType
    {
        typedef typename Type<ChildType> RET;
    };
};

namespace GenerateSubClassInternals
{

/**
* Consumes one type of the list and calls itself recursively to continue
* type usage.
*/
template <
    class Env_,         // environment to be passed to the classes
    class FinalType,    // The baseclass of the hierarchy
    class SubTypeList   // the list of types to add
         >
struct GenerateSubClass_
{
    typedef Env_ Env;
    typedef SubTypeList::UseType<GenerateSubClass_<Env, FinalType, SubTypeList::Tail>::RET >::RET RET;
};

/**
* Template specialization. Ends the recursion.
*/
template <
    class Env_,      // environment to be passed to the classes
    class FinalType  // The baseclass of the hierarchy
         >
struct GenerateSubClass_ <Env_, FinalType, EndTypeList >
{
    typedef Env_ Env;
    typedef FinalType RET;
};

} // end namespace GenerateSubClassInternals

/**
* Public template that receives a list of the hierarchy to be built and builds it.
*
* the Resulting type is available through the ResultType typedef.
*
* This works only when the BaseClass expects a structure like the following:
*
*   typedef Env: any information that needs to be passed to fill the parameters of the base class
*   typedef ResultType: final type of the subclass structure (this may or may not be used by the class)
*/
template <
    class Env_,                       // environment to be passed to the Base template
    template<class>class Base,        
    class SubTypeList = EndTypeList
         >
struct GenerateSubClass
{
    typedef Env_ Env;  
    typedef GenerateSubClassInternals::GenerateSubClass_<Env, Base<GenerateSubClass>, SubTypeList>::RET ResultType;
};


#endif //_SMP_INHERIT
